#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>

#define MAX_LENGTH 255
#define BAD_NUMBER {return '\0';}
#define GOOD_NUMBER return (char)(num + 33);
#define SPECIAL_UP (special == 1)
#define NUMBERS_UP (numbers == 1)
#define UPPER_UP (latin_upper == 1)
#define LOWER_UP (latin_lower == 1)

int get_passlen(void *arg);
int give_password(void *buf, int len);
char select_char(int num, int special, int latin_upper, int latin_lower, int numbers);
void generate_password(char *word, int len, int special, int latin_upper, int latin_lower, int numbers);
void set_flags(int num, char **keys, int *special, int *latin_upper, int *latin_lower, int *numbers);

int main(int argc, char **argv)
{
	int passlen;
	char password[MAX_LENGTH+2];
	int special = 0, latin_upper = 0, latin_lower = 0, numbers = 0;

	if (argc < 2) {
		printf("Please call passgen with length key\n");
		exit(1);
	} else if (argc == 2) {
		passlen = get_passlen(argv[1]);
		generate_password(password, passlen, 1, 1, 1, 1);
		give_password(password, passlen);
	} else  if (argc < 7){
		passlen = get_passlen(argv[1]);
		set_flags(argc, argv, &special, &latin_upper, &latin_lower, &numbers);
		generate_password(password, passlen, special, latin_upper, latin_lower, numbers);
		give_password(password, passlen);
	} else {
		printf("Too many keys\n");
		exit(1);
	}

	return 0;	
}

// function gets the length
int get_passlen(void *arg)
{
	int len;

	len = atoi(arg);

	if (len  == 0) {
		printf("Please enter the length as integer number as the first key\n");
		exit(1);
	}
	if (len > (MAX_LENGTH + 1) ) {
		printf("You entered too big length\n");
		exit(1);
	}

	return len;
}

// function returns a character
// see ascii table to understand it
char select_char(int num, int special, int latin_upper, int latin_lower, int numbers)
{
	if (num < 0) BAD_NUMBER
	else if (num < 15) {

		if SPECIAL_UP GOOD_NUMBER else BAD_NUMBER
	
	} else if (num < 25) {

		if NUMBERS_UP GOOD_NUMBER else BAD_NUMBER
	
	} else if (num < 32) {

		if SPECIAL_UP GOOD_NUMBER else BAD_NUMBER

	} else if (num < 58) {

		if UPPER_UP GOOD_NUMBER else BAD_NUMBER

	} else if (num < 64) {

		if SPECIAL_UP GOOD_NUMBER else BAD_NUMBER

	} else if (num < 90) {

		if LOWER_UP GOOD_NUMBER else BAD_NUMBER

	}else if (num < 94) {

		if SPECIAL_UP GOOD_NUMBER else BAD_NUMBER

	} else BAD_NUMBER
}

// function generating password
void generate_password(char *word, int len, int special, int latin_upper, int latin_lower, int numbers)
{
	int fd, i, res;
	unsigned short seeding[3];
	char newchar;

	// /dev/random - random number generator, read the random(4)
	fd = open("/dev/random", O_RDONLY);
	if (fd == -1) {
		printf("Random isn't initialized\n");
		exit(1);
	}

	//fill seeding with random bits from /dev/random
	res = read(fd, seeding, 3);
	if (res == -1) {
		printf("Seeding wasn't extracted\n");
		exit(1);
	}

	//collect the seeding - lrand48(3) returns random numbers
	seed48(seeding);

	//fill each character password with random character
	for (i = 0; i < len; i++) {
		do {
			res = lrand48() % 94;
			newchar = select_char(res, special, latin_upper, latin_lower, numbers);
		} while (newchar == '\0');
		word[i] = newchar;
	}

	close(fd);
}

// function sending password to stdout
int give_password(void *buf, int len) 
{
	int res, left;
	left = len;
	
	while (left) {
		res = fwrite(buf, sizeof(char), left, stdout);
		if(res == -1)  {
			if (errno == EINTR)
				continue;
			return -1;
		}
		buf += res;
		left -= res;
	}
	return 0;
}

// function setting the flags for password generating
void set_flags(int num, char **keys, int *special, int *latin_upper, int *latin_lower, int *numbers)
{
	int i;

	for (i = 2; i < num; i++) {
		switch (keys[i][0]) {
		case 's':
			*special = 1;
			break;
		case 'L':
			*latin_upper = 1;
			break;
		case 'l':
			*latin_lower = 1;
			break;
		case 'n':
			*numbers = 1;
			break;
		default:
			printf("Please check entered keys\n");
			exit(1);
		}
	}
}
